// XmlResource.cpp

#include "stdafx.h"
#include "XmlResource.h"
#include "SAX.h"
#include <vector>
#include <algorithm>

namespace Xml
{
	namespace Private
	{
		// common
		const TCHAR* const INVALID_XML			= _T("XMLu%svǂݍ߂܂Bt@CȂA\Ɍ肪܂B");	// (LPCTSTR)filename
		const TCHAR* const INVALID_ELEMENT		= _T("Ggu%lsv𗝉ł܂B܂B");	// (LPCWSTR)pwchLocalName
		const TCHAR* const INVALID_COMMAND		= _T("R}hu%lsv݂͑܂B܂B");	// (LPCWSTR)cmd

		// menu
		static LPWSTR MENU_ROOT			= L"menu";
		static LPWSTR MENU_SUBMENU		= L"submenu";
		static LPWSTR MENU_ITEM			= L"item";
		static LPWSTR MENU_SEPARATOR	= L"separator";
		static LPWSTR MENU_TITLE		= L"title";
		static LPWSTR MENU_COMMAND		= L"command";
		const TCHAR* const MENU_HAS_NO_TITLE	= _T("menutitle܂B");
		const TCHAR* const ITEM_HAS_NO_COMMAND	= _T("itemcommand܂B܂B");
		const TCHAR* const ITEM_HAS_NO_TITLE 	= _T("itemtitle܂B܂B");
		const WCHAR* const DEFAULT_MENU_TITLE	= L"???";

		// accel
		static LPWSTR ACCEL_ROOT		= L"accelerators";
		static LPWSTR ACCEL_BIND		= L"bind";
		static LPWSTR ACCEL_MODIFIER	= L"modifier";
		static LPWSTR ACCEL_KEY			= L"key";
		static LPWSTR ACCEL_COMMAND		= L"command";
		const TCHAR* const BIND_HAS_NO_MODIFIER	= _T("bindmodifier܂B܂B");
		const TCHAR* const BIND_HAS_NO_KEY		= _T("bindkey܂B܂B");
		const TCHAR* const BIND_HAS_NO_COMMAND	= _T("bindcommand܂B܂B");
		const TCHAR* const INVALID_KEYNAME		= _T("L[u%lsv݂͑܂B܂B");	// (LPCWSTR)VirtualKeyName
		const TCHAR* const INVALID_MODIFIER		= _T("Cqu%lcv𗝉ł܂B܂B");	// (WCHAR)modifier
	}

	static inline bool wcseq(LPCWSTR lhs, LPCWSTR rhs)
	{
		return wcscmp(lhs, rhs) == 0;
	}

	static LPCWSTR GetValueFromName(ISAXAttributes* pAttr, LPWSTR attrName)
	{
		static LPWSTR EMPTY_ATTR = L"";
		LPWSTR value;
		int valueLen;
		if SUCCEEDED(pAttr->getValueFromName(EMPTY_ATTR, 0, attrName, (int)wcslen(attrName), &value, &valueLen))
			return value;
		else
			return NULL;
	}

	////////////////////////////////////////////////////////////
	// LoadMenu
	//

	class MenuHandler : public SAXContentHandler
	{
	private:
		std::vector<HMENU>		m_MenuStack;
		ILoadResourceHandler*	m_pHandler;
	public:
		MenuHandler(ILoadResourceHandler* pHandler) : m_pHandler(pHandler)
		{}
		~MenuHandler()
		{
			if(!m_MenuStack.empty())
				::DestroyMenu(m_MenuStack.front());
		}
		HMENU	GetResult()
		{
			if(m_MenuStack.empty())
				return NULL;
			HMENU hMenu = m_MenuStack.front();
			m_MenuStack.clear();
			return hMenu;
		}

		STDMETHODIMP startElement( 
			LPWSTR pwchNamespaceUri,	int cchNamespaceUri,
			LPWSTR pwchLocalName,		int cchLocalName,
			LPWSTR pwchRawName,			int cchRawName,
			ISAXAttributes* pAttributes)
		{
			if(wcseq(pwchLocalName, Private::MENU_ROOT))
			{
				if(!m_MenuStack.empty())
				{
					::DestroyMenu(m_MenuStack.front());
					m_MenuStack.clear();
				}
				m_MenuStack.push_back(::CreateMenu());
			}
			else if(wcseq(pwchLocalName, Private::MENU_SUBMENU))
			{
				LPCWSTR title = GetValueFromName(pAttributes, Private::MENU_TITLE);
				if(title == NULL)
				{
					m_pHandler->OnWarning(Private::MENU_HAS_NO_TITLE);
					title = Private::DEFAULT_MENU_TITLE;
				}
				HMENU hSubMenu = ::CreateMenu();
				::AppendMenu(m_MenuStack.back(), MF_POPUP, (UINT_PTR)hSubMenu, _bstr_t(title));
				m_MenuStack.push_back(hSubMenu);
			}
			else if(wcseq(pwchLocalName, Private::MENU_ITEM))
			{
				LPCWSTR title = GetValueFromName(pAttributes, Private::MENU_TITLE);
				if(title == NULL)
				{
					m_pHandler->OnWarning(Private::ITEM_HAS_NO_TITLE);
					return S_OK;
				}
				LPCWSTR command = GetValueFromName(pAttributes, Private::MENU_COMMAND);
				if(command == NULL)
				{
					m_pHandler->OnWarning(Private::ITEM_HAS_NO_COMMAND);
					return S_OK;
				}
				UINT id = m_pHandler->NameToID(command);
				if(id == 0)
				{
					TCHAR msg[1024];
					wsprintf(msg, Private::INVALID_COMMAND, command);
					m_pHandler->OnWarning(msg);
					return S_OK;
				}
				::AppendMenu(m_MenuStack.back(), MF_STRING, id, _bstr_t(title));
			}
			else if(wcseq(pwchLocalName, Private::MENU_SEPARATOR))
			{
				::AppendMenu(m_MenuStack.back(), MF_SEPARATOR, 0, 0);
			}
			else
			{
				TCHAR msg[1024];
				wsprintf(msg, Private::INVALID_ELEMENT, pwchLocalName);
				m_pHandler->OnWarning(msg);
			}
			return S_OK;
		}
		STDMETHODIMP endElement( 
			LPWSTR pwchNamespaceUri,	int cchNamespaceUri,
			LPWSTR pwchLocalName,		int cchLocalName,
			LPWSTR pwchRawName,			int cchRawName)
		{
			if(wcseq(pwchLocalName, Private::MENU_SUBMENU))
			{
				m_MenuStack.pop_back();
			}
			return S_OK;
		}
	};
	HMENU	LoadMenuFromFile(LPCTSTR filename, ILoadResourceHandler* pHandler)
	{
		MenuHandler handler(pHandler);
		HRESULT hr = ReadXmlFromFile(_bstr_t(filename), &handler);
		if FAILED(hr)
		{
			TCHAR msg[1024];
			wsprintf(msg, Private::INVALID_XML, filename);
			pHandler->OnError(msg);
			return NULL;
		}
		return handler.GetResult();
	}
	HMENU	LoadMenuFromStream(VARIANT src, ILoadResourceHandler* pHandler)
	{
		MenuHandler handler(pHandler);
		HRESULT hr = ReadXmlFromStream(src, &handler);
		if FAILED(hr)
		{
			TCHAR msg[1024];
			wsprintf(msg, Private::INVALID_XML, _T("VARIANT"));
			pHandler->OnError(msg);
			return NULL;
		}
		return handler.GetResult();
	}

	////////////////////////////////////////////////////////////
	// LoadAccelerators
	//

	struct VirtualKey
	{
		LPCWSTR		name;
		WORD		id;

		friend bool operator < (const VirtualKey& lhs, LPCWSTR rhs)
		{
			return wcsicmp(lhs.name, rhs) < 0;
		}
	};
	// L[ to L[R[h zB啶ʂȂL[Ń\[gς݂ł邱ƁB
	const VirtualKey VIRTUALKEYS[] =
	{
		L","			, 0xBC,
		L"-"			, 0xBD,
		L"."			, 0xBE,
		L"/"			, 0xBF,
		L"0"			, 0x30,
		L"1"			, 0x31,
		L"2"			, 0x32,
		L"3"			, 0x33,
		L"4"			, 0x34,
		L"5"			, 0x35,
		L"6"			, 0x36,
		L"7"			, 0x37,
		L"8"			, 0x38,
		L"9"			, 0x39,
		L":"			, 0xBA,
		L";"			, 0xBB,
		L"@"			, 0xC0,
		L"["			, 0xDB,
		L"\\"			, 0xDC,
		L"]"			, 0xDD,
		L"^"			, 0xDE,
		L"_"			, 0xE2,
		L"a"			, 0x41,
		L"add"			, VK_ADD,
		L"b"			, 0x42,
		L"back"			, VK_BACK,
		L"backspace"	, VK_BACK,
		L"bs"			, VK_BACK,
		L"c"			, 0x43,
		L"d"			, 0x44,
		L"decimal"		, VK_DECIMAL,
		L"delete"		, VK_DELETE,
		L"div"			, VK_DIVIDE,
		L"divide"		, VK_DIVIDE,
		L"down"			, VK_DOWN,
		L"e"			, 0x45,
		L"end"			, VK_END,
		L"enter"		, VK_RETURN,
		L"esc"			, VK_ESCAPE,
		L"escape"		, VK_ESCAPE,
		L"f"			, 0x46,
		L"f1"			, VK_F1,
		L"f10"			, VK_F10,
		L"f11"			, VK_F11,
		L"f12"			, VK_F12,
		L"f13"			, VK_F13,
		L"f14"			, VK_F14,
		L"f15"			, VK_F15,
		L"f16"			, VK_F16,
		L"f17"			, VK_F17,
		L"f18"			, VK_F18,
		L"f19"			, VK_F19,
		L"f2"			, VK_F2,
		L"f20"			, VK_F20,
		L"f21"			, VK_F21,
		L"f22"			, VK_F22,
		L"f23"			, VK_F23,
		L"f24"			, VK_F24,
		L"f3"			, VK_F3,
		L"f4"			, VK_F4,
		L"f5"			, VK_F5,
		L"f6"			, VK_F6,
		L"f7"			, VK_F7,
		L"f8"			, VK_F8,
		L"f9"			, VK_F9,
		L"g"			, 0x47,
		L"h"			, 0x48,
		L"home"			, VK_HOME,
		L"i"			, 0x49,
		L"ins"			, VK_INSERT,
		L"insert"		, VK_INSERT,
		L"j"			, 0x4A,
		L"k"			, 0x4B,
		L"l"			, 0x4C,
		L"left"			, VK_LEFT,
		L"m"			, 0x4D,
		L"mul"			, VK_MULTIPLY,
		L"multiply"		, VK_MULTIPLY,
		L"n"			, 0x4E,
		L"next"			, VK_NEXT,
		L"num *"		, VK_MULTIPLY,
		L"num +"		, VK_ADD,
		L"num -"		, VK_SUBTRACT,
		L"num ."		, VK_DECIMAL,
		L"num /"		, VK_DIVIDE,
		L"num 0"		, VK_NUMPAD0,
		L"num 1"		, VK_NUMPAD1,
		L"num 2"		, VK_NUMPAD2,
		L"num 3"		, VK_NUMPAD3,
		L"num 4"		, VK_NUMPAD4,
		L"num 5"		, VK_NUMPAD5,
		L"num 6"		, VK_NUMPAD6,
		L"num 7"		, VK_NUMPAD7,
		L"num 8"		, VK_NUMPAD8,
		L"num 9"		, VK_NUMPAD9,
		L"o"			, 0x4F,
		L"p"			, 0x50,
		L"prior"		, VK_PRIOR,
		L"q"			, 0x51,
		L"r"			, 0x52,
		L"return"		, VK_RETURN,
		L"right"		, VK_RIGHT,
		L"s"			, 0x53,
		L"sep"			, VK_SEPARATOR,
		L"separator"	, VK_SEPARATOR,
		L"space"		, VK_SPACE,
		L"sub"			, VK_SUBTRACT,
		L"subtract"		, VK_SUBTRACT,
		L"t"			, 0x54,
		L"tab"			, VK_TAB,
		L"u"			, 0x55,
		L"up"			, VK_UP,
		L"v"			, 0x56,
		L"w"			, 0x57,
		L"x"			, 0x59,
		L"y"			, 0x59,
		L"z"			, 0x5A,
		L""			, VK_LEFT,
		L""			, VK_UP,
		L""			, VK_RIGHT,
		L""			, VK_DOWN,
	};

	class AccelHandler : public SAXContentHandler
	{
	private:
		typedef	std::vector<ACCEL>	AccelList;
		AccelList		m_Accels;
		ILoadResourceHandler*	m_pHandler;
	public:
		AccelHandler(ILoadResourceHandler* pHandler) : m_pHandler(pHandler)
		{}
		~AccelHandler()
		{}
		HACCEL	GetResult()
		{
			HACCEL hAccel = ::CreateAcceleratorTable(&m_Accels[0], (int)m_Accels.size());
			m_Accels.clear();
			return hAccel;
		}

		STDMETHODIMP startElement( 
			LPWSTR pwchNamespaceUri,	int cchNamespaceUri,
			LPWSTR pwchLocalName,		int cchLocalName,
			LPWSTR pwchRawName,			int cchRawName,
			ISAXAttributes* pAttributes)
		{
			if(wcseq(pwchLocalName, Private::ACCEL_ROOT))
			{
				m_Accels.clear();
				m_Accels.reserve(20);
				return S_OK;
			}
			else if(wcseq(pwchLocalName, Private::ACCEL_BIND))
			{
				ACCEL accel;
				// Cq̏
				LPCWSTR modifier = GetValueFromName(pAttributes, Private::ACCEL_MODIFIER);
				if(modifier == NULL)
				{
					m_pHandler->OnWarning(Private::BIND_HAS_NO_MODIFIER);
					return S_OK;
				}
				accel.fVirt = AnalyzeModifier(modifier);
				// zL[R[h̏
				LPCWSTR key = GetValueFromName(pAttributes, Private::ACCEL_KEY);
				if(key == NULL)
				{
					m_pHandler->OnWarning(Private::BIND_HAS_NO_KEY);
					return S_OK;
				}
				WORD vkey = ToVirtualKey(key);
				if(vkey == 0)
				{
					TCHAR msg[1024];
					wsprintf(msg, Private::INVALID_KEYNAME, key);
					m_pHandler->OnWarning(msg);
					return S_OK;
				}
				accel.key = vkey;
				// R}h̏
				LPCWSTR command = GetValueFromName(pAttributes, Private::ACCEL_COMMAND);
				if(command == NULL)
				{
					m_pHandler->OnWarning(Private::BIND_HAS_NO_COMMAND);
					return S_OK;
				}
				UINT id = m_pHandler->NameToID(command);
				if(id == 0)
				{
					TCHAR msg[1024];
					wsprintf(msg, Private::INVALID_COMMAND, command);
					m_pHandler->OnWarning(msg);
					return S_OK;
				}
				accel.cmd = id;
				m_Accels.push_back(accel);
				return S_OK;
			}
			else
			{
				TCHAR msg[1024];
				wsprintf(msg, Private::INVALID_ELEMENT, pwchLocalName);
				m_pHandler->OnWarning(msg);
			}
			return S_OK;
		}
		static WORD ToVirtualKey(LPCWSTR name)
		{
			const VirtualKey* begin = VIRTUALKEYS;
			const VirtualKey* end   = begin + sizeof(VIRTUALKEYS) / sizeof(VIRTUALKEYS[0]);
			const VirtualKey* found	= std::lower_bound(begin, end, name);
			if(found != end && wcsicmp(found->name, name) == 0)
				return found->id;
			else
				return 0;
		}
		BYTE AnalyzeModifier(LPCWSTR m)
		{
			BYTE mod = FVIRTKEY;
			for(; *m != L'\0'; ++m)
			{
				switch(*m)
				{
				case 'a': case 'A':
					mod |= FALT;
					break;
				case 'c': case 'C':
					mod |= FCONTROL;
					break;
				case 's': case 'S':
					mod |= FSHIFT;
					break;
				default:
					TCHAR msg[1024];
					wsprintf(msg, Private::INVALID_MODIFIER, *m);
					m_pHandler->OnWarning(msg);
					break;
				}
			}
			return mod;
		}
	};
#ifdef _DEBUG
	static void VERIFY_VIRTUALKEYS()
	{
		// VIRTUALKEYS ̓\[gς݂łȂ΂ȂȂ̂ŁAꉞmFĂB
		const int NUMKEYS = sizeof(VIRTUALKEYS) / sizeof(VIRTUALKEYS[0]);
		for(int i = 0; i < NUMKEYS-1; i++)
		{
			if(!(VIRTUALKEYS[i] < VIRTUALKEYS[i+1].name))
			{
				LPCWSTR lhs = VIRTUALKEYS[i].name;
				LPCWSTR rhs = VIRTUALKEYS[i+1].name; 
				__debugbreak();
			}
		}
	}
#else
	inline static void VERIFY_VIRTUALKEYS()	{}
#endif
	HACCEL	LoadAcceleratorsFromFile(LPCTSTR filename, ILoadResourceHandler* pHandler)
	{
		VERIFY_VIRTUALKEYS();

		AccelHandler handler(pHandler);
		HRESULT hr = ReadXmlFromFile(_bstr_t(filename), &handler);
		if FAILED(hr)
		{
			TCHAR msg[1024];
			wsprintf(msg, Private::INVALID_XML, filename);
			pHandler->OnError(msg);
			return NULL;
		}
		return handler.GetResult();
	}
	HACCEL	LoadAcceleratorsFromStream(VARIANT src, ILoadResourceHandler* pHandler)
	{
		VERIFY_VIRTUALKEYS();

		AccelHandler handler(pHandler);
		HRESULT hr = ReadXmlFromStream(src, &handler);
		if FAILED(hr)
		{
			TCHAR msg[1024];
			wsprintf(msg, Private::INVALID_XML, _T("VARIANT"));
			pHandler->OnError(msg);
			return NULL;
		}
		return handler.GetResult();
	}
}
